"""# lucidium.version

Version tracking utility.
"""

__author__ = "Gabriel C. Trahan"
__author_email__ = "gabrieltrahan777@hotmail.com"
__description__ = """Experiments in neuro-symbolic reinforcement learning in the pursuit of developing agents/methods to aid in organic reasoning, inference, and logic."""
__title__ = "lucidium"
__url__ = "https://github.com/theokoles7/lucidium"
__version__ = "0.1.2"        
__version_info__ = tuple(int(i) for i in __version__.split("."))