"""# lucidium..symbolic

This package defines various structures needed for symbolic learning.
"""

__all__ =   [
                "Predicate"
            ]

from lucidium.symbolic.predicate    import Predicate