"""# lucidium

Experiments in neuro-symbolic reinforcement learning in the pursuit of developing agents/methods to 
aid in organic reasoning, inference, and logic.
"""

__all__ =   [
                "__author__",
                "__author_email__",
                "__description__",
                "__title__",
                "__url__",
                "__version__",
                "__version_info__",
            ]

from lucidium._version_ import  (
                                    __author__,
                                    __author_email__,
                                    __description__,
                                    __title__,
                                    __url__,
                                    __version__,
                                    __version_info__,
                                )